@@ -3,7 +3,7 @@
  * rum.h
  *	  Exported definitions for RUM index.
  *
- * Portions Copyright (c) 2015-2016, Postgres Professional
+ * Portions Copyright (c) 2015-2019, Postgres Professional
  * Portions Copyright (c) 2006-2016, PostgreSQL Global Development Group
  *
  *-------------------------------------------------------------------------
@@ -19,6 +19,8 @@
 #include "access/sdir.h"
 #include "lib/rbtree.h"
 #include "storage/bufmgr.h"
+#include "utils/datum.h"
+#include "utils/memutils.h"
 
 #include "rumsort.h"
 
@@ -262,6 +264,16 @@ typedef signed char RumNullCategory;
 /*
  * Data (posting tree) pages
  */
+/*
+ * FIXME -- Currently RumItem is placed as a pages right bound and PostingItem
+ * is placed as a non-leaf pages item. Both RumItem and PostingItem stores
+ * AddInfo as a raw Datum, which is bogus. It is fine for pass-by-value
+ * attributes, but it isn't for pass-by-reference, which may have variable
+ * length of data. This AddInfo is used only by order_by_attach indexes, so it
+ * isn't allowed to create index using ordering over pass-by-reference AddInfo,
+ * see initRumState(). This can be solved by having non-fixed length right bound
+ * and non-fixed non-leaf posting tree item.
+ */
 #define RumDataPageGetRightBound(page)	((RumItem*) PageGetContents(page))
 #define RumDataPageGetData(page)	\
 	(PageGetContents(page) + MAXALIGN(sizeof(RumItem)))
@@ -519,7 +531,7 @@ extern void rumEntryFillRoot(RumBtree btree, Buffer root, Buffer lbuf, Buffer rb
 				 Page page, Page lpage, Page rpage);
 extern IndexTuple rumPageGetLinkItup(RumBtree btree, Buffer buf, Page page);
 extern void rumReadTuple(RumState * rumstate, OffsetNumber attnum,
-			 IndexTuple itup, RumItem * items);
+			 IndexTuple itup, RumItem * items, bool copyAddInfo);
 extern void rumReadTuplePointers(RumState * rumstate, OffsetNumber attnum,
 					 IndexTuple itup, ItemPointerData *ipd);
 extern void updateItemIndexes(Page page, OffsetNumber attnum, RumState * rumstate);
@@ -773,8 +785,8 @@ extern IndexBulkDeleteResult *rumvacuumcleanup(IndexVacuumInfo *info,
 extern bool rumvalidate(Oid opclassoid);
 
 /* rumbulk.c */
-#if PG_VERSION_NUM <= 100006
-#define RBTNode RBNode
+#if PG_VERSION_NUM <= 100006 || PG_VERSION_NUM == 110000
+typedef RBNode RBTNode;
 #endif
 
 typedef struct RumEntryAccumulator
@@ -931,10 +943,14 @@ rumDataPageLeafReadItemPointer(char *ptr, ItemPointer iptr, bool *addInfoIsNull)
  * Reads next item pointer and additional information from leaf data page.
  * Replaces current item pointer with the next one. Zero item pointer should be
  * passed in order to read the first item pointer.
+ *
+ * It is necessary to pass copyAddInfo=true if additional information is used
+ * when the data page is unlocked. If the additional information is used without
+ * locking one can get unexpected behaviour.
  */
 static inline Pointer
 rumDataPageLeafRead(Pointer ptr, OffsetNumber attnum, RumItem * item,
-					RumState * rumstate)
+					bool copyAddInfo, RumState * rumstate)
 {
 	Form_pg_attribute attr;
 
@@ -999,8 +1015,13 @@ rumDataPageLeafRead(Pointer ptr, OffsetNumber attnum, RumItem * item,
 		}
 		else
 		{
-			ptr = (Pointer) att_align_pointer(ptr, attr->attalign, attr->attlen, ptr);
-			item->addInfo = fetch_att(ptr, attr->attbyval, attr->attlen);
+			Datum		addInfo;
+
+			ptr = (Pointer) att_align_pointer(ptr, attr->attalign, attr->attlen,
+											  ptr);
+			addInfo = fetch_att(ptr, attr->attbyval, attr->attlen);
+			item->addInfo = copyAddInfo ?
+				datumCopy(addInfo, attr->attbyval, attr->attlen) : addInfo;
 		}
 
 		ptr = (Pointer) att_addlength_pointer(ptr, attr->attlen, ptr);
@@ -1062,7 +1083,10 @@ extern Datum FunctionCall10Coll(FmgrInfo *flinfo, Oid collation,
 				   Datum arg9, Datum arg10);
 
 /* PostgreSQL version-agnostic creation of memory context */
-#if PG_VERSION_NUM >= 110000
+#if PG_VERSION_NUM >= 120000
+#define RumContextCreate(parent, name) \
+	AllocSetContextCreate(parent, name, ALLOCSET_DEFAULT_SIZES)
+#elif PG_VERSION_NUM >= 110000
 	#define RumContextCreate(parent, name) \
 		AllocSetContextCreateExtended(parent, name, \
 									  ALLOCSET_DEFAULT_MINSIZE, \
