@@ -3,7 +3,7 @@
  * rum_arr_utils.c
  *		various anyarray-search functions
  *
- * Portions Copyright (c) 2015-2016, Postgres Professional
+ * Portions Copyright (c) 2015-2019, Postgres Professional
  * Portions Copyright (c) 1996-2016, PostgreSQL Global Development Group
  *
  *-------------------------------------------------------------------------
@@ -51,7 +51,11 @@
 
 #define CHECKARRVALID(x) \
 	do { \
-		if (x) { \
+		if (x == NULL) \
+			ereport(ERROR, \
+					(errcode(ERRCODE_NULL_VALUE_NOT_ALLOWED), \
+					 errmsg("array must not be NULL"))); \
+		else if (x) { \
 			if (ARR_NDIM(x) != NDIM && ARR_NDIM(x) != 0) \
 				ereport(ERROR, \
 						(errcode(ERRCODE_ARRAY_SUBSCRIPT_ERROR), \
@@ -733,6 +737,7 @@ cmpAscArrayElem(const void *a, const void *b, void *arg)
 {
 	FmgrInfo	*cmpFunc = (FmgrInfo*)arg;
 
+	Assert(a && b);
 	return DatumGetInt32(FunctionCall2Coll(cmpFunc, DEFAULT_COLLATION_OID, *(Datum*)a, *(Datum*)b));
 }
 
